import random
import subprocess

path_to_AutoMoDe_executable = "/home/jkuckling/Documents/MasterThesis/2020_VincentVanPelt/AutoMoDe-BT/bin/automode_main_bt"
# scenario_file = "/home/jkuckling/Documents/MasterThesis/2020_VincentVanPelt/Cedrata_extensions/missions/foraging_bt.argos"

class BehaviorTreeNode():

    def __init__(self):
        pass

    def to_cmd(self, id="root"):
        pass

class ControlFlowNode(BehaviorTreeNode):

    def __init__(self, children):
        self.children = children

    def build_args(self, node_id, type_id):
        args = ["--n{}".format(node_id), type_id,
                "--nchild{}".format(node_id), str(len(self.children))]
        if node_id == "root":
            base_id = ""
        else:
            base_id = node_id
        for count, child in enumerate(self.children):
            args.extend(child.to_cmd(base_id + str(count)))
        return args

class SelectorNode(ControlFlowNode):

    def __init__(self, children):
        self.children = children

    def to_cmd(self, id="root"):
        return self.build_args(id, "0")


class SelectorStarNode(ControlFlowNode):

    def __init__(self, children):
        self.children = children

    def to_cmd(self, id="root"):
        return self.build_args(id, "1")


class SequenceNode(ControlFlowNode):

    def __init__(self, children):
        self.children = children

    def to_cmd(self, id="root"):
        return self.build_args(id, "2")


class SequenceStarNode(ControlFlowNode):

    def __init__(self, children):
        self.children = children

    def to_cmd(self, id="root"):
        return self.build_args(id, "3")


class ActionNode(BehaviorTreeNode):

    def __init__(self, action, params):
        self.action_int = action
        self.params = params

    def to_cmd(self, identifier):
        i = identifier
        args = ["--n{}".format(i), "5", "--a{}".format(i), str(self.action_int)]
        for param in self.params:
            if param == "att" or param == "rep" or param == "dmin":
                pval = "%.2f" % self.params[param]
            elif param == "rwm" or param == "sig" or param == "tmax" or param == "min" or param == "max":
                pval = str(self.params[param])
            else:
                print("Undefined parameter {}".format(param))
                pval = 0
            args.extend(["--{}{}".format(param, i), pval])
        args.extend(["--p{}".format(i), "0"])
        return args


class ConditionNode(BehaviorTreeNode):

    def __init__(self, condition, params):
        self.condition_int = condition
        self.params = params

    def to_cmd(self, identifier):
        i = identifier
        args = ["--n{}".format(i), "6", "--c{}".format(i), str(self.condition_int)]
        for param in self.params:
            # "BlackFloor", "GrayFloor", "WhiteFloor", "FixedProbability"
            if self.condition_int == 0 or self.condition_int == 1 or self.condition_int == 2 or self.condition_int == 5:
                if param=="p":
                    pval = "%.2f" % self.params[param]
            # "NeighborsCount", "InvertedNeighborsCount"
            if self.condition_int == 3 or self.condition_int == 4:
                if param == "w":
                    pval = "%.2f" % self.params[param]
                if param == "p":
                    pval = str(self.params[param])
            if self.condition_int == 9:
                if param == "sig":
                    pval = str(self.params[param])
            args.extend(["--{}{}".format(param, i), pval])
        return args


def check_if_controller_valid(controller):
    actions = []
    conditions = []
    nodes = [controller]
    while nodes:
        n = nodes.pop(0)
        if type(n) is ActionNode:
            actions.append(n)
        elif type(n) is ConditionNode:
            conditions.append(n)
        else:
            nodes.extend(n.children)
    if len(actions) > 4:
        return False
    if len(conditions) > 4:
        return False
    if any([a.params["min"] >= a.params["max"] for a in actions if a.action_int in [9, 10]]):
        return False
    return True

def evaluate_controller(controller, scenario_file):
    # create 10 random seeds
    seeds = [random.randint(0, 2147483647) for _ in range(0, 10)]
    if not check_if_controller_valid(controller):
        print("Invalid controller")
        return [-100] * 10
    # convert the controller to cmd args
    controller_cmd = controller.to_cmd()
    scores = []
    for seed in seeds:
        # prepare cmd
        args = [path_to_AutoMoDe_executable, "-n", "-c", scenario_file, "--seed", str(seed), "--bt-config"]
        args.extend(controller_cmd)
        p = subprocess.Popen(args, stderr=subprocess.PIPE, stdout=subprocess.PIPE)
        (stdout, stderr) = p.communicate()
        # Analyse result
        output = stdout.decode('utf-8')
        lines = output.splitlines()
        try:
            # print(lines[len(lines) - 1])
            score = float(lines[len(lines) - 1].split(" ")[1])
        except:
            score = -1  # Just to be sure
            print("Args: " + str(args))
            print("Stderr: " + stderr.decode('utf-8'))
            print("Stdout: " + stdout.decode('utf-8'))
            raise
        scores.append(score)
    return scores
